"""Useful decorators.
"""

import pprint
import os
import json
import uuid
import logging as rawLogger
import datetime
import time
import threading
import functools

from contextlib import ContextDecorator

import wrapt

from flask import request, g

DEFAULT_LOGGER = rawLogger.getLogger(__name__)


def _start_watch(name, w_type, args, kwargs, show_args=False,
                 tag=None, logger=DEFAULT_LOGGER):
    """Helper function to start watching command

    :param name:    String name of command to watch.

    :param w_type:  String type (e.g., 'function', 'flask_request').

    :param args:    Arguments to command.

    :param kwargs:  Keyword args.

    :param show_args=False:  Whether to show args/kwargs in log.

    :param tag:  Optional string or callable which takes in *args, **kwargs
                 and produces a string. This can be used to generate a tag
                 based on the args and kwargs and get saved in the logs
                 was 'w_tag'.

    :param logger:  Optional logger to use.

    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    :return:  Dictionary of meta data for watched command.

    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    PURPOSE:  Create dictionary of meta data and log it so we can
              watch a command. Meant to be called by watched decorator.

    """

    w_data = {'w_uuid': str(uuid.uuid4()),  # UUID to help tracking.
              'w_name': name, 'watched': 'start', 'w_type': w_type,
              'start': time.time()}
    if show_args:
        w_data.update(w_args=args, w_kwargs=str(kwargs))
    if isinstance(tag, str):
        w_data['w_tag'] = tag
    elif callable(tag):
        w_data['w_tag'] = tag(*args, **kwargs)

    logger.info('watched_cmd: %s', name, extra=w_data)
    return w_data


def _end_watch(w_data, str_result, logger=DEFAULT_LOGGER):
    """Log an error in a watched command.

    :param w_data:    As produced by _start_watch.

    :param str_result:  Result of command as a string.

    :param logger:  Optional logger to use.

    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    :return:  Updated `w_data`.

    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    PURPOSE:  Log about a watched command ending.
              Helper meant to be called only be `watched` decorator.

    """

    cmd_time = time.time() - w_data['start']
    w_data.update(watched='end', status='ok', w_run_time=cmd_time,
                  w_result=str_result)
    logger.info('watched_cmd_end: ok:%s (%.4f s)', w_data['w_name'],
                 cmd_time, extra=w_data)
    return w_data


def _error_watch(w_data, my_problem, logger=DEFAULT_LOGGER):
    """Log an error in a watched command.

    :param w_data:    As produced by _start_watch.

    :param my_problem:    Problem description.

    :param logger:  Optional logger to use.

    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    :return:  Updated `w_data`.

    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    PURPOSE:  Log a warning about an error in a watched command.
              Helper meant to be called only be `watched` decorator.

    """
    cmd_time = time.time() - w_data['start']
    w_data.update(w_error=str(my_problem), status='error',
                  w_run_time=cmd_time, watched='end')
    logger.warning('watched_end_cmd: error:%s', w_data['w_name'],
                    extra=w_data)
    return w_data


def watched(wrapped=None, show_args=True, tag=None, logger=DEFAULT_LOGGER):
    """Decorator to make a command "watched" where we track timing.

    :param wrapped:    Function to wrap.

    :param show_args=True:  Whether to show args in logs.

    :param tag:  Optional string or callable which takes in *args, **kwargs
                 and produces a string. This can be used to generate a tag
                 based on the args and kwargs and get saved in the logs
                 was 'w_tag'. Note that the tag function will not knowing
                 anything about default argument values so you will need
                 to define those in the tag function as well.

    :param logger:  Optional logger to use.

    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    :return:  Same as calling wrapped function.

    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    PURPOSE:  Decorator to make a "watched" function where we put info
              about start, end, error, args, into logs. This will create
              log messages starting with 'watched_*' and providing an
              `extra` dictionary to logger with useful meta data. You can
              then search your logs for things like 'watched_end_cmd' to
              find info on how the command ran.

              This is especially useful with something like logz.io.

Below we illustrate example usage

>>> from ox_ui.core import decorators
>>> @decorators.watched  # use decorator to create a watched function
... def foo(x, y=1):
...     'example function to watch'
...     return f'x={x}, y={y}'
...
>>> foo(2)  # Nothing special happens unless you turn on logging at INFO level
'x=2, y=1'

We can also customize the logger used (e.g., to make the log messages easier
to see or test) and create a custom tag for each function call:

>>> @decorators.watched(
...     logger=decorators.FakeLogger(),                  # use custom logger
...     tag=lambda x, y='unknown': f'tag: {x % 2}, {y}'  # create custom tag
... )
... def bar(x, y=1):
...     'example function to watch'
...     return f'x={x}, y={y}, x+y={x+y}'
...
>>> bar(3)  # doctest: +ELLIPSIS,+NORMALIZE_WHITESPACE
DEBUG: watched_meta: Ignoring instance None
INFO: watched_cmd: bar
  extra={'start': ...
 'w_args': (3,),
 'w_kwargs': '{}',
 'w_name': 'bar',
 'w_tag': 'tag: 1, unknown',
 'w_type': 'function',
 'w_uuid': '...',
 'watched': 'start'}
INFO: watched_cmd_end: ok:bar (... s)
  extra={'start': ...,
 'status': 'ok',
 'w_args': (3,),
 'w_kwargs': '{}',
 'w_name': 'bar',
 'w_result': 'x=3, y=1, x+y=4',
 'w_run_time': ...,
 'w_tag': 'tag: 1, unknown',
 'w_type': 'function',
 'w_uuid': '...',
 'watched': 'end'}
'x=3, y=1, x+y=4'

    """
    if wrapped is None:  # triggered when decorator called with arguments
        return functools.partial(
            watched, show_args=show_args, tag=tag, logger=logger)

    @wrapt.decorator
    def outer_wrapper(wrapped, instance, args, kwargs):
        logger.debug('watched_meta: Ignoring instance %s', instance)
        name = wrapped.__name__
        w_data = _start_watch(name, 'function', args, kwargs, show_args,
                              tag=tag, logger=logger)
        try:
            str_result = '(unknown)'
            result = wrapped(*args, **kwargs)
            try:
                str_result = str(result)
                if len(str_result) > 200:
                    str_result = str_result[:195]+'...'
            except Exception as unexpected:  # pylint: disable=broad-except
                logger.error('Ignoring unexpected str conversion exception: %s',
                              unexpected)
            _end_watch(w_data, str_result, logger=logger)
            return result
        except Exception as my_problem:  # pylint: disable=broad-except
            _error_watch(w_data, my_problem, logger=logger)
            raise
    return outer_wrapper(wrapped)


def _start_watching_request(req=None, name=None):
    """Start watching a flask request.

Uses the `_start_watch`, `_end_watch` helpers to watch a flask
command. Use `setup_flask_watch` to turn on for your app. See
docs for `setup_flask_watch` for details.
    """
    req = req or request
    name = name or request.endpoint
    w_data = _start_watch(name, 'flask_request', None, None, False)
    g.w_data = w_data  # pylint: disable=assigning-non-slot


def _end_watching_request(req=None, logger=DEFAULT_LOGGER):
    """End watching a flask request.

Uses the `_start_watch`, `_end_watch` helpers to watch a flask
command. Use `setup_flask_watch` to turn on for your app. See
docs for `setup_flask_watch` for details.
    """
    req = req or request
    w_data = getattr(g, 'w_data', None)
    if w_data is None:
        logger.error('No w_data member for request: %s', req)
        return
    str_result = ''
    _end_watch(w_data, str_result[:195])


@watched
def setup_flask_watch(app):
    """Take a flask app as input and setup logs to watch requests.

This uses tools similar to the `watched` decorator to watch and log
start/end of a flask command using before_request and teardown_request
hooks.

Basically just call this on your app and then search your logs for
something like `watched_cmd*` to see info about your flask commands.
    """
    app.before_request(_start_watching_request)
    app.teardown_request(_end_watching_request)


class LockFile(ContextDecorator):
    """Context decorator to create/check lock files.

This class can serve as a decorator or a context manager to
create/check a lock file.

The following illustrates example usage:

>>> import time, threading, os, logging
>>> from ox_ui.core import decorators
>>> lock_path = '/tmp/test.lock'     # example lock file path
>>> @decorators.LockFile(lock_path)  # decorate function to use lock file
... def foo(t, m):
...     assert os.path.exists(lock_path)
...     logging.warning('%s Sleep %s thread %s', m, t, threading.get_ident())
...     time.sleep(t)
...
>>> my_thread = threading.Thread(target=foo, args=[5, 'thread'])
>>> problems = []  # Thread is started in background.
>>> try:           # Verify a FileExistsError raised if conflict occurs.
...     my_thread.start()  # start a background thread with the lock
...     while not os.path.exists(lock_path):
...         time.sleep(0.5)
...     foo(4, 'main')     # now spawn in main thread to poke lock
... except Exception as problem:
...     problems.append(problem)
...
>>> assert len(problems) > 0, problems  # Verify we caught an exception.

    """

    def __init__(self, lockpath, comment='', encoding='utf8'):
        self.lockpath = lockpath
        self.encoding = encoding
        self.comment = comment
        self.created = None

    def remove_lock(self):
        """Remove lock if it exists.
        """
        if os.path.exists(self.lockpath):
            os.remove(self.lockpath)

    def __enter__(self):
        if os.path.exists(self.lockpath):
            info = 'unknown'
            try:
                with open(self.lockpath, encoding=self.encoding) as fdesc:
                    info = json.load(fdesc)
                DEFAULT_LOGGER.warning('Found lock file %s with data: %s',
                                       self.lockpath, info)
            except Exception:  # pylint: disable=broad-except
                DEFAULT_LOGGER.exception(
                    'Unable to get info about lock file %s', self.lockpath)
            raise FileExistsError(self.lockpath)
        with open(self.lockpath, 'w', encoding=self.encoding) as fdesc:
            self.created = datetime.datetime.now()
            info = {'pid': os.getpid(), 'comment': self.comment,
                    'thread_id': threading.get_ident(),
                    'created_dt': str(self.created),
                    'created_ts': self.created.timestamp()}
            json.dump(info, fdesc, indent=2)
        return self

    def __exit__(self, exc_type, exc, exc_tb):
        if exc_type:
            DEFAULT_LOGGER.debug('Ignoring exception info %s, %s, %s',
                                 exc_type, exc, exc_tb)
        self.remove_lock()
        return False


class FakeLogger:
    """Fake logging object to echo log messages to stdout for tests.

You can use this class to test logging as shown below:

The following just uses a FakeLogger to make the log messages go to stdout:

>>> from ox_ui.core import decorators
>>> fl = decorators.FakeLogger()
>>> fl.debug('hi %s', 'you')
DEBUG: hi you
>>> fl.info('hi %s', 'there')
INFO: hi there

The following just uses a FakeLogger to collect log messages:

>>> log = []  # log messages will get saved to a list
>>> fl = decorators.FakeLogger(echo=log.append)
>>> fl.debug('hi %s', 'you')
>>> fl.info('hi %s', 'there')
>>> print(log)  # show the accumulated log messages
['DEBUG: hi you', 'INFO: hi there']
    """

    def __init__(self, echo=print):
        self._echo = echo
        for name in ['debug', 'info', 'warning', 'error', 'exception']:
            setattr(self, name, functools.partial(self._show, name))

    def _show(self, level, msg, *args, extra=None):
        p_msg = msg % args
        full_msg = f'{level.upper()}: {p_msg}'
        if extra:
            full_msg += f'\n  extra={pprint.pformat(extra)}'
        self._echo(full_msg)



"""Tools to convert click commands to flask WTForms
"""
import typing
import datetime
import tempfile
import re
import logging
import pathlib
import os
import warnings
from io import BytesIO, StringIO
from contextlib import contextmanager

from flask import render_template, make_response, request
from flask_wtf import FlaskForm
from jinja2 import Environment, BaseLoader
from click import types, utils
from wtforms import widgets
from wtforms import (
    StringField, IntegerField, FileField, Field, PasswordField, SelectField,
    BooleanField, HiddenField, FloatField)
from wtforms.validators import DataRequired
from dateutil.parser import parse, ParserError

from ox_ui import core as ox_ui_core
from ox_ui.assets import css


@contextmanager
def ensure_open(
        filelike: typing.Union[str, utils.LazyFile, BytesIO, StringIO],
        *args, **kwargs):
    """
    ensure `filelike` is opened and ready for read or write.

    `filelike` could be a str showing a file path,
    or an instance of click.File, or BytesIO and StringIO in case the
    field is gobbled.

    For the first two cases, we will open `filelike`,
    the mode will be specified in `*args` and `**kwargs`
    if `filelike` is a str, or implied inside `filelike` if it is a click.File.

    No need to open BytesIO or StringIO.
    We must NOT close BytesIO or StringIO and leave it to client code to close
    after `getvalue`. Otherwise, their content will be discarded right after
    close.
    """
    opened = None
    try:
        if isinstance(filelike, str):
            opened = open(filelike, *args, **kwargs)
        elif isinstance(filelike, utils.LazyFile):
            opened = filelike.open()
        yield opened or filelike
    finally:
        if opened:
            opened.close()


class RequestResult(typing.NamedTuple):
    """Result of handle_request method of ClickToWTF class.
    """

    finished: bool
    form: typing.Any
    output: typing.Any


class FileResCallback:
    """
    to replace the deprecated FileResponseTweak
    """
    def __init__(self, arg_name, mode, filename):
        self.arg_name = arg_name
        self.mode = mode
        self.filename = filename
        self.buffer = None

    def gobble(self, cmd, name):
        dummy = cmd
        return name == self.arg_name

    def pad_kwargs(self, cmd, kwargs):
        dummy = cmd
        self.buffer = BytesIO() if 'b' in self.mode else StringIO()
        kwargs[self.arg_name] = self.buffer

    def name(self):
        return self.__class__.__name__

    def post_process_result(self, cmd, result):
        logging.debug('Tweak %s ignores previous result of %s', self.name(),
                      result)

        try:
            bytes_ = self.buffer.getvalue()
            if isinstance(bytes_, str):
                bytes_ = bytes_.encode('utf-8')
        finally:
            self.buffer.close()
        response = make_response(bytes_)
        response.headers.set('Content-Type', 'application/octest-stream')
        response.headers.set(
            'Content-Disposition', 'attachment', filename=self.filename)
        return response


class FileResponseTweak:
    def __init__(self, arg_name, split_char='_', **mktemp_kwargs):
        self.arg_name = arg_name
        self.split_char = split_char
        self.mktemp_kwargs = dict(mktemp_kwargs)
        self.file_name = None
        warnings.warn(
            'FileResponseTweak is deprecated. Use FileResCallback instead',
            DeprecationWarning)

    def gobble(self, cmd, name):
        dummy = cmd
        return name == self.arg_name

    def pad_kwargs(self, cmd, kwargs):
        dummy = cmd
        self.file_name = tempfile.mktemp(**self.mktemp_kwargs)
        kwargs[self.arg_name] = self.file_name

    def name(self):
        return self.__class__.__name__

    def post_process_result(self, cmd, result):
        logging.debug('Tweak %s ignores previous result of %s', self.name(),
                      result)
        response = None
        # Use binary mode below to get consistent behaviour across platforms
        with open(self.file_name, 'rb') as my_fd:
            suffix = self.mktemp_kwargs.get('suffix', None)
            if suffix and self.split_char and self.split_char in suffix:
                fname = self.file_name.split(self.split_char)[-1]
            else:
                fname = self.file_name
            response = make_response(my_fd.read())
            response.headers.set('Content-Type', 'application/octest-stream')
            response.headers.set(
                'Content-Disposition', 'attachment', filename=fname)

        if os.path.exists(self.file_name):
            logging.info('Removing temporary file %s', self.file_name)
            os.remove(self.file_name)
        return response


class DateTimeFieldTweak(Field):
    """Ttext field which stores a `datetime.datetime` from multiple formats.
    """
    widget = widgets.TextInput()

    def __init__(self, label=None, validators=None,
                 formats=('%Y-%m-%d %H:%M:%S', '%Y-%m-%d'), **kwargs):
        super(DateTimeFieldTweak, self).__init__(label, validators, **kwargs)
        self.formats = formats

    def _value(self):
        if self.raw_data:
            return ' '.join(self.raw_data)
        if self.data:
            if isinstance(self.data, str):
                return self.data
            return self.data.strftime(self.formats[0])
        return ''

    def process_formdata(self, valuelist):
        if valuelist:
            if date_str := ' '.join(valuelist):
                # populate self.data with data_str to pass DataRequired check,
                # so we show parse error instead of the "required field" error
                # on bad inputs
                self.data = date_str
                try:
                    self.data = parse(date_str)
                except ParserError as err:
                    raise ValueError(str(err))


class ClickToWTF:

    def __init__(self, clickCmd, skip_opt_re=None, tweaks: list = None,
                 fill_get_from_url: bool = False,
                 hidden_overrides = None,
                 require_re = None):
        """Initializer.

        :param clickCmd:   A click command to turn into a form.

        :param skip_opt_re=None:  Regexp for options to skip in form.

        :param tweaks:  List of 'tweaks' to add (e.g., FileResponseTweak).

        :param fill_get_from_url:  If True, then when doing a GET request
                                   to fill out arguments we look in the
                                   URL params to override standard defaults.

        :param hidden_overrides:  Optional dictionary with string keys
                                  indicating fields to hide in WTForm and
                                  values indicating the value to provide.
                                  For example, if you provide something like
                                  {'user': 'anon'} then the option 'user' in
                                  `clickCmd` will not be visibile in the
                                  WTForm but contain a hidden value set to
                                  `'anon'`. This is useful to force certain
                                  values in the web form.
        """
        self.clickCmd = clickCmd
        self.rawTemplate = None
        self.skip_opt_re = skip_opt_re if not skip_opt_re else re.compile(
            skip_opt_re)
        self.tweaks = tweaks if tweaks else []
        self.fill_get_from_url = fill_get_from_url
        self.gobbled_opts = {}
        self.hidden_overrides = dict(
            hidden_overrides) if hidden_overrides else {}
        self.require_re = dict(require_re) if require_re else {}

    def click_cmd_params(self):
        """
        to make sure skip_opt_re and gobbled specs are respected
        """
        for opt in self.clickCmd.params:
            if self.skip_opt_re and self.skip_opt_re.search(opt.name):
                logging.info('Option %s since matchs skip_opt_re', opt.name)
            elif self.gobble(opt.name):
                logging.info('Option %s gobbled', opt.name)
            else:
                yield opt

    def form_cls(self):

        class ClickForm(FlaskForm):
            """Form to run Click command.
            """

        for opt in self.click_cmd_params():
            field = self.click_opt_to_wtf_field(opt)
            setattr(ClickForm, opt.name, field)

        if request.method == 'GET' and self.fill_get_from_url:
            self.populate_form_from_url(ClickForm)

        for name, value in self.hidden_overrides.items():
            setattr(ClickForm, name, HiddenField(name, default=value))

        return ClickForm

    def populate_form_from_url(self, ClickForm):
        """Populate a form from URL parameters.

        :param ClickForm:  Class form click form as produced by `form_cls`.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:  Go through parameters for command and try to get default
                  values from URL parameters and put them into `ClickForm`.
                  This is useful for dynamically sending users to a command
                  with special default values.

                  This is mainly intended to be called from form_cls
                  only if self.fill_get_from_url is True because it is
                  a minor security risk. A malicious actor could send
                  the user to the form with malicious defaults so we only
                  do this if fill_get_from_url is set to True or if this is
                  explicitly called by a "safe" command.
        """
        logging.debug('Populating form %s from URL', ClickForm)
        for opt in self.click_cmd_params():
            url_value = request.args.get(opt.name, None)
            if url_value is not None:
                field = getattr(ClickForm, opt.name)
                field.kwargs['default'] = url_value

    def gobble(self, name):
        for tweak in self.tweaks:
            reason = tweak.gobble(self, name)
            if reason:
                self.gobbled_opts[name] = reason
                return reason
        return None

    def pad_kwargs(self, kwargs):
        for tweak in self.tweaks:
            tweak.pad_kwargs(self, kwargs)

    def form(self):
        try:
            cls = self.form_cls()
        except Exception as prob:  # pylint: disable=broad-except
            #  Sometimes if you do a POST and do not provide all the
            #  values expected by a form, newer versions of Flask will
            #  have problems.
            logging.exception(
                'Unable to make form. If doing a POST; verify data provided.')
            raise
        return cls()

    @classmethod
    def click_opt_to_wtf_field(cls, opt):
        validators = []
        if opt.required:
            validators.append(
                DataRequired(f'* required field'))
        kwargs = dict(
            validators=validators,
            description=str(opt.help),
            default=opt.default)

        if opt.type == types.INT:
            return IntegerField(opt.name, **kwargs)
        if opt.type == types.FLOAT:
            return FloatField(opt.name, **kwargs)        
        if opt.type == types.BOOL:
            return BooleanField(opt.name, **kwargs)
        if opt.type == types.STRING:
            field_cls = PasswordField if opt.hide_input else StringField
            return field_cls(opt.name, **kwargs)
        if isinstance(opt.type, types.DateTime) or (
                getattr(opt.type, 'name', '?') == 'datetime'):
            if hasattr(opt, 'formats'):
                kwargs['formats'] = opt.formats
            default = opt.default
            if callable(default):
                default = default()
            if isinstance(default, str):
                default = datetime.datetime.strptime(
                    default, '%Y-%m-%d %H:%M:%S')
            kwargs['default'] = default

            return DateTimeFieldTweak(opt.name, **kwargs)
        if isinstance(opt.type, types.File):
            if opt.type.mode in ('r', 'rb'):  # file to read
                return FileField(opt.name, **kwargs)
        if getattr(opt.type, 'name') == 'choice':
            return cls.handle_choice_type(opt, **kwargs)
        raise TypeError(f'Cannot represent click type {opt.type} in WTF')

    @staticmethod
    def handle_choice_type(opt, **kwargs):
        """Turn click.Choice into SelectField.
        """
        return SelectField(opt.name, choices=[
            (n, n) for n in opt.type.choices], **kwargs)

    def _verify_kwargs(self, kwargs):
        for name, value in self.hidden_overrides.items():
            actual = kwargs.get(name, None)
            regexp = self.require_re.get(name, None)
            if regexp is not None:
                pass  # we will validate later when we check require_re
            elif actual != value:
                #  Someone may have manually tried to circumvent override
                raise ValueError(f'Field {name} must match default {value}'
                                 f' (got {actual}).')
        for name, value in self.require_re.items():
            actual = kwargs.get(name, '')
            if not re.compile(value).match(actual):
                raise ValueError(
                    f'Field {name} does not match regexp {value}'
                    f' (got {actual}).')

    def process(self, form):
        defaults = {opt.name: opt.default for opt in self.clickCmd.params}
        kwargs = {}
        for opt in self.click_cmd_params():
            value = getattr(form, opt.name).data
            if opt.multiple:
                kwargs[opt.name] = [*kwargs.get(opt.name, []), value]
            else:
                kwargs[opt.name] = value
        kwargs = {**defaults, **kwargs}
        self._verify_kwargs(kwargs)
        self.pad_kwargs(kwargs)
        wrapped_cmd = getattr(self.clickCmd.callback, '__wrapped__', None)
        if wrapped_cmd:
            raw_result = wrapped_cmd(**kwargs)
        else:
            raw_result = self.clickCmd.callback(**kwargs)

        return self.post_process(raw_result)

    def post_process(self, raw_result):
        result = raw_result
        for tweak in self.tweaks:
            result = tweak.post_process_result(self, result)
        return result

    def get_form_data(self) -> typing.Dict[str, typing.Any]:
        """Return dictionary of form data for show_form method.

This method returns a dictionary with string keys that will be used by
the `show_form` method in rendering the form for the command.

The following are provided by default:

  - intro:  Text from the `help` variable of the command.

A standard pattern is to override this via something like

        def get_form_data(self):
            data = super().get_form_data()
            data['my_custom_data'] = SOMETHING
            return data

        """
        return {'intro': self.clickCmd.help}

    def show_form(self, form=None, template=None):
        if form is None:
            form = self.form()
        data = self.get_form_data()
        if template:
            result = render_template(template, form=form, **data)
        else:
            result = self.simple_render(form, **data)
        return result

    def simple_render(self, form, **data):
        if not self.rawTemplate:
            tpath = pathlib.Path(ox_ui_core.__file__).parent.joinpath(
                'sample_wtf.html')
            cpath = pathlib.Path(css.__file__).parent.joinpath('w3.css')
            self.rawTemplate = '<style>\n%s</style>\n%s\n' % (
                open(cpath).read(), open(tpath).read())
        rtemplate = Environment(loader=BaseLoader()).from_string(
            self.rawTemplate)
        result = rtemplate.render(form=form, **data)
        return result

    def handle_request(self, verbose=False):
        form = self.form()
        if form.validate_on_submit():
            result = RequestResult(True, form, self.process(form))
        else:
            logging.debug('Showing form either because invalid or first view')
            result = RequestResult(False, form, self.show_form(form=form))
        if verbose:
            return result
        return result.output



"""Tools to help test servers.
"""

import pathlib
import logging
import subprocess
import socket


class ServerInfo:
    """Simple class to hold basic server information.
    """

    def __init__(self, s_proc, s_port):
        self.s_proc = s_proc
        self.s_port = s_port


def find_free_port():
    "Find and return a free port number"

    result = None
    my_sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        my_sock.bind(('', 0))
        my_sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        result = my_sock.getsockname()[1]
    finally:
        my_sock.close()
    return result


def run_cmd(cmd, cwd=None, timeout=30, env=None):
    """Helper to run a command in a subprocess.

    :param cmd:   List of strings for commands to run.

    :param cwd=None:    Optional current working directory string.

    :param timeout=30:  Timeout in seconds before killing command.

    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    :return:  The subprocess created to run the command.

    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    PURPOSE:  Provides some convenience for callnig subprocess.Popen
              to run subprocess commands.

    """
    logging.info('Running cmd: %s', str(cmd))
    proc = subprocess.Popen(cmd, cwd=cwd, env=env)
    if timeout is not None:
        try:
            result = proc.wait(timeout=abs(timeout))
        except subprocess.TimeoutExpired:
            if timeout < 0:
                result = proc.returncode  # timeout is OK
            else:
                raise  # timeout not ok so re-raise it
    else:
        result = proc.returncode
    if result:
        raise ValueError('Error code "%s" in cmd "%s"' % (
            result, proc))
    return proc


def start_server(cmd: list, cwd: str = None, s_port=None, env=None):
    """Start flask server.

    :param cmd:   Command to run to start server.

    :param cwd:      Current working directory to run in.

    :param s_port=None:    Optional port for server. If None, we choose.

    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    :return:  Instance of ServerInfo containing information about the
              server we started.

    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    PURPOSE:  Use python subprocess module to start an instance of the
              server in the background. You can use the returned
              ServerInfo instance to interact with the server. This is
              useful for making the test self-contained.

              You can call kill_server to cleanup or restart_server
              if you want to restart during interactive testing.
    """
    s_port = s_port if s_port else str(find_free_port())
    cwd = cwd if cwd else pathlib.Path(cmd[0]).parent
    cmd.extend(['--port', s_port])
    server = run_cmd(cmd=cmd, cwd=cwd, timeout=-4, env=env)
    return ServerInfo(server, int(s_port))



"""A setuptools based setup module for ox_ui

Copyright (c) 2016--2019, Emin Martinian - All Rights Reserved
See LICENSE at the top-level of this distribution for more information
or write to emin.martinian@gmail.com.
"""

# see also setup.cfg

from os import path

from setuptools import setup, find_packages
from ox_ui import VERSION


def get_readme():
    'Get the long description from the README file'

    here = path.abspath(path.dirname(__file__))
    # README.rst is autogenerated from README.org via something like
    # pandoc --from=org --to=rst --output=README.rst README.org
    with open(path.join(here, 'README.rst'), encoding='utf-8') as my_fd:
        result = my_fd.read()

    return result


setup(
    name='ox_ui',
    version=VERSION,
    description='Simple user interface tools for python',
    long_description=get_readme(),
    url='http://github.com/emin63/ox_ui',
    author='Emin Martinian',
    author_email='emin.martinian@gmail.com',
    license='custom',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3',
    ],


    keywords='flask click ui',
    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),
    include_package_data=True,
    install_requires=['pytest'],
    # If there are data files included in your packages that need to be
    # installed, specify them here.
    package_data={
        'sample': ['package_data.dat'],
    },
    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    entry_points={
        'console_scripts': [
            'sample=sample:main',
        ],
    },
)
